//////////////////////////////////////////////////////////////////////
//
//  XBPGET modified to work with AK_SAY() and AK_GET()
//
//  Copyright:
//     Ashok Kumar
//  
//  Contents:
//     Clipper @..SAY..GET.. implementation with Picture and VALID
//  
//  Remarks:
//     Converts Clipper's @..SAY.. into XBPStatics and
//     Converts Clipper's @..GET.. into XBPSle
//  
//  Return:
//     returns nil
//  
//  History:
//
//   8 May   98 - DISABLE GETS - similar to the original CLEAR GETS
//  21 April 98 - SETKEY handler
//  17 April 98 - Fixes to XBPStatic
//  16 April 98 - make GetActive() work
//              - Validation on DATES - Test CENTURY and EPOCH
//              - Post translated data to the screen before leaving (dates etc)
//              - make SET CONFIRM to work - auto skip to next get
//              - new HIDDEN clause for passwords
//  15 April 98 - make VALID use the original Clipper conditions
//              - Get the RANGE clause working
//              - Tidyup XbpGet.CH
//              - BIG ONE - get the WHEN clause working
//              - Redo length calculations for gets using transform/picture
//  14 April 98 - Fixes to reset GET focus after VALID function calls
//              - Comment out Locking system for the time being
//  13 April 98 - Get VALID working
//  12 April 98 - It works !!!
//
//////////////////////////////////////////////////////////////////////
//
//  XBPGET.PRG
//
//  Copyright:
//     Alaska Software Inc., (c) 1997. All rights reserved.         
//  
//  Contents:
//     XbpGet implementation.
//  
//  Remarks:
//     Sample implementation of a XbpGet class derived from XbpSle which uses
//     a overloaded TextMode Get object as server for the picture and buffer
//     management.
//     This class demonstrates the use of the XbpSle:AutoKeyboard flag
//     and the use of the GET class.
//  
//     Note:
//     When validation is necessary, please refer to the validation 
//     mechanisms of the DataRef class.
//  
//  
//  Return:
//     returns 0 always 
//  
//////////////////////////////////////////////////////////////////////

#include "Appevent.ch"
#include "Common.ch"
#include "Gra.ch"
#include "Xbp.ch"
#include "Font.ch"


* **********************************************************************
* InvisibleGet service class
* **********************************************************************
CLASS InvisibleGet FROM Get
   EXPORTED:
   METHOD display
ENDCLASS

METHOD InvisibleGet:display()
RETURN self


* **********************************************************************
* XbpGet class
* **********************************************************************
CLASS XbpGet FROM XbpSLE

   // AK 16/4/98 - Export the Get as well
   *** HIDDEN:
   ***
   *** VAR    Get                     // the embedded invisible service GET

   EXPORTED:

   // AK 16/4/98 - Export the Get as well
   VAR    Get                     // the embedded invisible service GET
   VAR    GetNext

                                  ** creation flags
   VAR    Picture                 // my picture

   // AK 15/4/98
   VAR    When                    // when

                                  ** overloaded from XbpSle
   METHOD Init, Create            // Lifecycle
   METHOD Keyboard                // overloaded keyboard handling
   METHOD LbUp                    // to inquire mark

   METHOD CutMarked, PasteMarked, DeleteMarked // manipulate mark

                                  ** overloaded from DataRef
   METHOD SetData, GetData        // overloaded SetData, GetData
   METHOD Undo                    // overloaded Undo to Get

   METHOD SetEditBuffer           // change edit buffer

   // AK 16/4/98
   METHOD GetNext, GetPrev        // Jump to next get

ENDCLASS

* **********************************************************************
* Init method
* **********************************************************************
METHOD XbpGet:Init( oParent, oOwner, aPos, aSize, aPP, lVisible, nRow, nCol, cFldName )
   // AK 16/4/98 init parameters

   DEFAULT aPP TO {}

   * initialisation of base class
   AAdd( aPP, { XBP_PP_COMPOUNDNAME, FONT_DEFFIXED_SMALL } )
   ::XbpSle:init( oParent, oOwner, aPos, aSize, aPP, lVisible )

RETURN self

* **********************************************************************
* Create method
* **********************************************************************
METHOD XbpGet:Create( oParent, oOwner, aPos, aSize, aPP, lVisible, nRow, nCol, cFldName )
   // AK 16/4/98 create parameters

   * create the SLE and disable keyboard handling
   ::XbpSle:autoKeyboard := .F.
   ::XbpSle:Create( oParent, oOwner, aPos, aSize, aPP, lVisible )

   * now the datalink must have been specified
   IF ! ValType( ::DataLink ) == "B"
      ::Status := XBP_STAT_FAILURE
      RETURN self
   ENDIF

   * now initialize the get object
   // Add preBlock AK 15/4/98
   // Add row, col, fieldname AK 16/4/98
   ::Get := InvisibleGet():New(nRow, nCol, ::Datalink, cFldName, ::Picture, , ,::When)
   ::Get:SetFocus()
   ::BufferLength := Len( ::Get:buffer )
   IF ::Get:Type == "N"
      ::Align := XBPSLE_RIGHT                  // not supported on WIN32 ???
   ENDIF

   // disable if ! preBlock AK 15/4/98
   if Valtype(::Get:preBlock) == "B" .and. ! Eval( ::Get:preBlock, ::Get )
      ::disable()
   endif

RETURN self

* **********************************************************************
* Overloaded SetData()
*
* This method should not behave like the normal XbpSle:SetData() method
* which needs strings. This method will rather call the codeblock and
* use its return value and convert this value to the given datatype.
*
* Note: it is currently not allowed to assign any value to the XbpGet
*       directly
* **********************************************************************
METHOD XbpGet:SetData()

   * call updatebuffer on GET, this will call the datalink and set
   * the :Get:buffer value
   ::Get:UpdateBuffer()

   * now set the data in the SLE class
   ::SetEditBuffer( ::Get:Buffer )

   * now set the initial mark
   IF Set( _SET_INSERT )
      ::SetMarked ( { ::Get:Pos, ::Get:Pos } )
   ELSE
      ::SetMarked ( { ::Get:Pos, ::Get:Pos + 1 } )
   ENDIF

RETURN self

* **********************************************************************
* Overloaded GetData()
*
* Return untransformed contents of the GET buffer
* **********************************************************************
METHOD XbpGet:GetData()

   LOCAL xValue := ::Get:Untransform()

   * store value to datalink
   Eval( ::DataLink, xValue )

RETURN xValue

* **********************************************************************
* Overloaded Undo()
* **********************************************************************
METHOD XbpGet:Undo()

   ::Get:Undo()
   ::Get:Buffer := Transform( ::Get:Original, ::Get:Picture )

RETURN ::SetEditBuffer( ::Get:buffer )

* **********************************************************************
* Change the edit buffer of the XbpSle.
* **********************************************************************
METHOD XbpGet:SetEditBuffer( cBuffer )

   ::XbpSle:SetData( cBuffer )

RETURN self

* **********************************************************************
* Overloaded Keyboard handling
* **********************************************************************
METHOD XbpGet:Keyboard( nKey )
   LOCAL aMarked, cChar, lHandled := .T.
   LOCAL setPos    := .T.
   // AK 14/4/98
   LOCAL k:=i:=AScan(getlist, oXbp ), nGetLLen := Len(getlist), j, newKey
 
   aMarked := ::QueryMarked()

   // Check XBase Keycodes
   IF ( bkeyblock := SETKEY( nKey ) ) != nil
      EVAL( bkeyblock, PROCNAME( 2 ), PROCLINE( 2 ), oXbp:Get:Name )
      SetAppFocus( getlist[i] )
      RETURN self
   ENDIF

   // Now check Clipper Keycodes
   newKey := key_trans(nKey)
   IF ( bkeyblock := SETKEY( newKey ) ) != nil
      EVAL( bkeyblock, PROCNAME( 2 ), PROCLINE( 2 ), oXbp:Get:Name )
      SetAppFocus( getlist[i] )
      RETURN self
   ENDIF

   DO CASE

   // Jump forward using Enter, Tab or Down Arrow    AK 13/4/98
   CASE nKey == xbeK_RETURN .or. nKey == xbeK_DOWN .or. nKey == xbeK_TAB
      ::GetNext()

   // Jump backwards using Shift Tab or Up Arrow    AK 13/4/98
   CASE nKey == xbeK_UP .or. nKey == xbeK_SH_TAB
      ::GetPrev()

   // Exit using Esc, F10, PageUP or PageDown    AK 13/4/98
   CASE nKey == xbeK_ESC .or. nKey == xbeK_F10 .or. ;
        nKey == xbeK_PGDN .or. nKey == xbeK_PGUP .or. nkey == xbeK_CTRL_W
        nKey := 27
        ::Undo()
        ::getData()
        PostAppEvent( xbeP_Close,,, self)

   CASE nKey == xbeK_SH_RIGHT
        ::Get:Pos ++
        aMarked[2] := ::Get:Pos
        ::SetMarked( aMarked )
        setPos := .F.

   CASE nKey == xbeK_SH_LEFT
        ::Get:Pos --
        aMarked[1] := ::Get:Pos
        ::SetMarked( aMarked )
        setPos := .F.

   CASE nKey == xbeK_SH_END
        aMarked[2] := ::Bufferlength + 1
        ::Get:Pos  := ::Bufferlength
        ::SetMarked( aMarked )
        setPos := .F.

   CASE nKey == xbeK_SH_HOME
        ::Get:Pos  := 1
        aMarked[1] := 1
        ::SetMarked( aMarked )
        setPos := .F.
         
   CASE nKey == xbeK_CTRL_U
        ::Undo()

   CASE nKey == xbeK_RIGHT
        ::Get:Right()

   CASE nKey == xbeK_CTRL_RIGHT
        ::Get:WordRight()

   CASE nKey == xbeK_LEFT
        ::Get:Left()

   CASE nKey == xbeK_CTRL_LEFT
        ::Get:WordLeft()

   CASE nKey == xbeK_HOME
        ::Get:Home()
        ::SetFirstChar( 1 )

   CASE nKey == xbeK_END
        ::Get:_End()

   CASE nKey == xbeK_INS
        Set( _SET_INSERT, ! Set( _SET_INSERT) )

   CASE nKey == xbeK_SH_INS
        ::PasteMarked()

   CASE nKey == xbeK_CTRL_INS
        ::CopyMarked()

   CASE nKey == xbeK_BS
        ::Get:BackSpace()

   CASE nKey == xbeK_DEL
        ::DeleteMarked()

   CASE nKey == xbeK_SH_DEL
        ::CutMarked()

   CASE nKey == xbeK_CTRL_T
        ::Get:DelWordRight()

   CASE nKey == xbeK_CTRL_Y
        ::Get:DelEnd()

   CASE nKey == xbeK_CTRL_BS
        ::Get:DelWordLeft()

   CASE nKey >= 32 .AND. nKey <= 255
      cChar := Chr(nKey)

      IF ::Get:Type == "N" .AND. cChar $ ".,"
         ::Get:ToDecPos()
      ELSE

         IF Set(_SET_INSERT)
            ::Get:Insert( cChar )
         ELSE
            ::Get:Overstrike( cChar )
         ENDIF

         // AK 16/4/98 - Exit if no space remaining and CONFIRM is OFF
         IF ::Get:typeOut .and. ! Set(_SET_CONFIRM)
            ::GetNext()
         ENDIF

      ENDIF

   OTHERWISE
     lHandled := .F.

   ENDCASE

   IF ! ::EditBuffer() == ::Get:Buffer
      ::SetEditBuffer( ::Get:Buffer )
   ENDIF

   IF setPos
      IF Set( _SET_INSERT )
         aMarked := { ::Get:Pos, ::Get:Pos }
      ELSE
         aMarked := { ::Get:Pos, ::Get:Pos + 1 }
      ENDIF
   ENDIF
   ::SetMarked( aMarked )

   IF ! lHandled
      RETURN ::XbpSle:Keyboard( nKey )
   ENDIF

RETURN self

* **********************************************************************
* Overloaded LbUp
* **********************************************************************
METHOD XbpGet:LbUp(mp1, mp2)

   LOCAL aMarked := ::queryMarked()

   * now locate Get on the specified position
   ::Get:Pos := aMarked[1]

   * now set the mark when insert mode is not active
   IF aMarked[2] - aMarked[1] == 0 .AND. ! Set( _SET_INSERT )
      ::SetMarked( { aMarked[1], aMarked[1]+1 } )
   ENDIF

RETURN ::XbpSle:LbUp( mp1, mp2 )

* **********************************************************************
* Overloaded CutMarked()
* **********************************************************************
METHOD XbpGet:CutMarked()

   ::CopyMarked()
   ::DeleteMarked()

RETURN self

* **********************************************************************
* Overloaded PasteMarked()
* **********************************************************************
METHOD XbpGet:PasteMarked()

   LOCAL aMarked := ::QueryMarked()
   LOCAL cString, oClipboard, i, nPaste

   * get string from clipboard
   oClipboard := XbpClipboard():New():Create()
   oClipboard:Open()
   cString := oClipboard:GetBuffer( XBPCLPBRD_TEXT )
   oClipboard:Close()

   * paste an empty string into the edit buffer when the clipboard
   * does not contain a valid string
   IF cString == NIL
      cString := ""
   ENDIF

   * clear the selected string and insert the current contents of the 
   * clipboard
   ::Get:Pos := aMarked[1]
   IF aMarked[2] - aMarked[1] <= 1
      nPaste := Len( cString )
   ELSE
      nPaste := Min( Len( cString ), aMarked[2] - aMarked[1] )
   ENDIF
   FOR i := 1 TO nPaste
      IF i <= Len( cString )
         ::Get:Overstrike( SubStr( cString, i, 1 ) )
      ELSE
         ::Get:Overstrike( " " )
      ENDIF
   NEXT

RETURN self

* **********************************************************************
* Overloaded DeleteMarked()
* **********************************************************************
METHOD XbpGet:DeleteMarked()

   LOCAL aMarked := ::QueryMarked()
   LOCAL i

   * delete the current selected string
   IF aMarked[2] - aMarked[1] <= 1
      ::Get:Delete()
   ELSE
      ::Get:Pos := aMarked[2]
      FOR i := 1 TO aMarked[2] - aMarked[1]
         ::Get:BackSpace()
      NEXT
   ENDIF

RETURN self

// EOF

* **********************************************************************
* New GetNext()     AK 16/4/98
* **********************************************************************
METHOD XbpGet:GetNext()
   LOCAL k:=i:=AScan(getlist, oXbp ), nGetLLen := Len(getlist), j

   ::getData()
   ::Get:Home()
   if ::Validate() .and. .not.(::Get:badDate())
      ::setData()
      IF ++i > nGetLLen
         PostAppEvent( xbeP_Close,,, self)
      ELSE
         // Prevalidation of gets
         FOR j := 1 TO Len(getlist)
            if Valtype(getlist[j]:Get:preBlock) == "B"
               if Eval( getlist[j]:Get:preBlock, getlist[j]:Get )
                  getlist[j]:enable()
               else
                  getlist[j]:disable()
               endif
            endif
         NEXT
         i := AK_GetNext(i, .F.)
         If i > nGetLLen
            PostAppEvent( xbeP_Close,,, self)
         else
            SetAppFocus( getlist[i] )
         endif
      ENDIF
   else
      SetAppFocus( getlist[i] )
      ::Undo()
      ::getData()
   endif
Return nil

* **********************************************************************
* New GetPrev()     AK 16/4/98
* **********************************************************************
METHOD XbpGet:GetPrev()
   LOCAL k:=i:=AScan(getlist, oXbp ), nGetLLen := Len(getlist), j

   ::getData()
   ::Get:Home()
   If ::Validate() .and. .not.(::Get:badDate())
      ::setData()
      i --
      IF i < 1
         PostAppEvent( xbeP_Close,,, self)
      ELSE
         // Prevalidation of gets
         FOR j := 1 TO Len(getlist)
            if Valtype(getlist[j]:Get:preBlock) == "B"
               if Eval( getlist[j]:Get:preBlock, getlist[j]:Get )
                  getlist[j]:enable()
               else
                  getlist[j]:disable()
               endif
            endif
         NEXT
         i := AK_GetNext(i, .T.)
         If i < 1
            PostAppEvent( xbeP_Close,,, self)
         else
            SetAppFocus( getlist[i] )
         endif
      ENDIF
   else
      SetAppFocus( getlist[i] )
      ::Undo()
      ::getData()
   Endif
Return Nil


//////////////////////////////////////////////////////////////////////
//
//  AK_SAY()
//
//  Copyright:
//     Ashok Kumar
//  
//  Contents:
//     Clipper @..SAY implementation.
//  
//  Remarks:
//     Converts Clipper's @..SAY into XBPStatics
//  
//  Return:
//     returns nil
//  
//////////////////////////////////////////////////////////////////////
Function AK_Say(nX, nY, cField, cPicture)
   LOCAL nFontHeight := SetAppWindow():fontHeight
   LOCAL nFontWidth  := 8       //SetAppWindow():fontWidth
   LOCAL aPos        := { (nY)*nFontWidth, (MaxRow()-nX)*nFontHeight }
   LOCAL nLength     := 0, nAtPos := 0, cSubst := ""

   // This code leaves the Says as hybrid screen output
   // Comment/delete if you want to use XBPStatics.
   if empty(cPicture)
      DevPos( nX, nY) ; DevOut( (cField) )
   else
      DevPos( nX, nY) ; DevOutPict( (cField), cPicture )
   endif
   Return nil
   //-------------------------------------------------

   // The rest of the code creates XBPStatics for each Say
   if empty(cPicture)
      cPicture := "@X"
      if ValType(cField) = "N"
         cPicture := "@N"
      elseif ValType(cField) = "D"
         cPicture := "@D"
      elseif ValType(cField) = "L"
         cPicture := "@Y"
      endif
      nLength := Len(Transform( cField, cPicture ))
   else
      nAtPos := at("@S", cPicture)
      if nAtPos > 0
         cSubst := substr(cPicture, nAtPos+2 )
         if at(" ", cSubst ) > 0
            nLength := Val( cSubst, at(" ", cSubst ) )
         else
            nLength := Val( cSubst )
         endif
      else
         nLength := Len(Transform( cField, cPicture ))
      endif
   endif
   aSize       := { (nLength+1)*nFontWidth, nFontHeight*1.1}

   oXbp         := XbpStatic():new(,, aPos, aSize )
   oXbp:type    := XBPSTATIC_TYPE_TEXT
   oXbp:options := XBPSTATIC_TEXT_VCENTER    //+ XBPSTATIC_TEXT_CENTER
   if ValType(cField) = "N"
      oXbp:caption := Str(cField)
   else
      oXbp:caption := cField
   endif
   oXbp:create()
   oXbp:setColorBG( GRA_CLR_PALEGRAY )
   AAdd( saylist, oXbp)

   // Reposition if x and y are row() and col()
   DevPos( nX, nY + nlength)

Return nil

//////////////////////////////////////////////////////////////////////
//
//  AK_GET()
//
//  Copyright:
//     Ashok Kumar
//  
//  Contents:
//     Clipper @..GET implementation with Picture and VALID
//  
//  Remarks:
//     Converts Clipper's @..GET into XBPSle
//  
//  Return:
//     returns nil
//  
//////////////////////////////////////////////////////////////////////
Function AK_Get(nX, nY, cField, cPicture, lValid, lWhen, lhidden)
   LOCAL nFontHeight := SetAppWindow():fontHeight
   LOCAL nFontWidth  := 8       //SetAppWindow():fontWidth
   LOCAL aPos        := { (nY)*nFontWidth, (MaxRow()-nX)*nFontHeight }
   LOCAL nLength     := 0, nAtPos := 0, cSubst := ""

   if empty(cPicture)
      cPicture := "@X"
      if ValType(&cField) = "N"
         cPicture := "@N"
      elseif ValType(&cField) = "D"
         cPicture := "@D"
      elseif ValType(&cField) = "L"
         cPicture := "@Y"
      endif
      nLength := Len(Transform( &cField, cPicture ))
   else
      nAtPos := at("@S", cPicture)
      if nAtPos > 0
         cSubst := substr(cPicture, nAtPos+2 )
         if at(" ", cSubst ) > 0
            nLength := Val( cSubst, at(" ", cSubst ) )
         else
            nLength := Val( cSubst )
         endif
      else
         nLength := Len(Transform( &cField, cPicture ))
      endif
   endif
   aSize       := { (nLength+1)*nFontWidth, nFontHeight*1.1}

   oXbp:= XbpGET():new( , , aPos, aSize, , , nX, nY, &cField )
   oXbp:clipSiblings := .T.
   oXbp:tabStop := .T.
   oXbp:autoTab := .T.
   oXbp:autoSize := .F.

   // Code for locking, updating and unlocking field by field
   oXbp:Datalink := { |x| IF( PCOUNT()==1, ( RLOCK(), &cField := x, dbUnlock() ), &cField ) }

   // No locking
   // oXbp:Datalink := { |x| IF( PCOUNT()==1,  &cField := x , &cField ) }

   oXbp:Picture  := cPicture
   if lValid <> NIL
      oXbp:Validate := &( "{|x|" + lValid + "}" )
   endif
   if lWhen <> NIL
      oXbp:When := &( "{|x|" + lWhen + "}" )
   endif

   if lhidden = NIL
      oXbp:unReadable := .F.
   else
      oXbp:unReadable:= .T.
   endif

   oXbp:create( , , aPos, aSize, , , nX, nY, cField )
   AAdd( getlist, oXbp)

Return nil

//////////////////////////////////////////////////////////////////////

Function AK_GetNext(glist, greverse)
   If greverse
      DO WHILE glist > 0 .and. ! getlist[glist]:isenabled()
         glist --
      ENDDO
   ELSE
      DO WHILE glist <= Len(getlist) .and. ! getlist[glist]:isenabled()
         glist ++
      ENDDO
   ENDIF
Return glist
//////////////////////////////////////////////////////////////////////

Function key_trans(nkey)
#include "inkey.ch"
   KeyCodes := { ;
    {xbeK_CTRL_END      ,  K_CTRL_W       }, ;
    {xbeK_CTRL_W        ,  K_CTRL_W       }, ;
    {xbeK_ALT_A         ,  K_ALT_A        }, ;
    {xbeK_ALT_B         ,  K_ALT_B        }, ;
    {xbeK_ALT_C         ,  K_ALT_C        }, ;
    {xbeK_ALT_D         ,  K_ALT_D        }, ;
    {xbeK_ALT_E         ,  K_ALT_E        }, ;
    {xbeK_ALT_F         ,  K_ALT_F        }, ;
    {xbeK_ALT_G         ,  K_ALT_G        }, ;
    {xbeK_ALT_H         ,  K_ALT_H        }, ;
    {xbeK_ALT_I         ,  K_ALT_I        }, ;
    {xbeK_ALT_J         ,  K_ALT_J        }, ;
    {xbeK_ALT_K         ,  K_ALT_K        }, ;
    {xbeK_ALT_L         ,  K_ALT_L        }, ;
    {xbeK_ALT_M         ,  K_ALT_M        }, ;
    {xbeK_ALT_N         ,  K_ALT_N        }, ;
    {xbeK_ALT_O         ,  K_ALT_O        }, ;
    {xbeK_ALT_P         ,  K_ALT_P        }, ;
    {xbeK_ALT_Q         ,  K_ALT_Q        }, ;
    {xbeK_ALT_R         ,  K_ALT_R        }, ;
    {xbeK_ALT_S         ,  K_ALT_S        }, ;
    {xbeK_ALT_T         ,  K_ALT_T        }, ;
    {xbeK_ALT_U         ,  K_ALT_U        }, ;
    {xbeK_ALT_V         ,  K_ALT_V        }, ;
    {xbeK_ALT_W         ,  K_ALT_W        }, ;
    {xbeK_ALT_X         ,  K_ALT_X        }, ;
    {xbeK_ALT_Y         ,  K_ALT_Y        }, ;
    {xbeK_ALT_Z         ,  K_ALT_Z        }, ;
    {xbeK_ALT_1         ,  K_ALT_1        }, ;
    {xbeK_ALT_2         ,  K_ALT_2        }, ;
    {xbeK_ALT_3         ,  K_ALT_3        }, ;
    {xbeK_ALT_4         ,  K_ALT_4        }, ;
    {xbeK_ALT_5         ,  K_ALT_5        }, ;
    {xbeK_ALT_6         ,  K_ALT_6        }, ;
    {xbeK_ALT_7         ,  K_ALT_7        }, ;
    {xbeK_ALT_8         ,  K_ALT_8        }, ;
    {xbeK_ALT_9         ,  K_ALT_9        }, ;
    {xbeK_ALT_0         ,  K_ALT_0        }, ;
    {xbeK_F1            ,  K_F1           }, ;
    {xbeK_F2            ,  K_F2           }, ;
    {xbeK_F3            ,  K_F3           }, ;
    {xbeK_F4            ,  K_F4           }, ;
    {xbeK_F5            ,  K_F5           }, ;
    {xbeK_F6            ,  K_F6           }, ;
    {xbeK_F7            ,  K_F7           }, ;
    {xbeK_F8            ,  K_F8           }, ;
    {xbeK_F9            ,  K_F9           }, ;
    {xbeK_F10           ,  K_F10          }, ;
    {xbeK_F11           ,  K_F11          }, ;
    {xbeK_F12           ,  K_F12          }, ;
    {xbeK_CTRL_F1       ,  K_CTRL_F1      }, ;
    {xbeK_CTRL_F2       ,  K_CTRL_F2      }, ;
    {xbeK_CTRL_F3       ,  K_CTRL_F3      }, ;
    {xbeK_CTRL_F4       ,  K_CTRL_F4      }, ;
    {xbeK_CTRL_F5       ,  K_CTRL_F5      }, ;
    {xbeK_CTRL_F6       ,  K_CTRL_F6      }, ;
    {xbeK_CTRL_F7       ,  K_CTRL_F7      }, ;
    {xbeK_CTRL_F8       ,  K_CTRL_F8      }, ;
    {xbeK_CTRL_F9       ,  K_CTRL_F9      }, ;
    {xbeK_CTRL_F10      ,  K_CTRL_F10     }, ;
    {xbeK_CTRL_F11      ,  K_CTRL_F11     }, ;
    {xbeK_CTRL_F12      ,  K_CTRL_F12     }, ;
    {xbeK_ALT_F1        ,  K_ALT_F1       }, ;
    {xbeK_ALT_F2        ,  K_ALT_F2       }, ;
    {xbeK_ALT_F3        ,  K_ALT_F3       }, ;
    {xbeK_ALT_F4        ,  K_ALT_F4       }, ;
    {xbeK_ALT_F5        ,  K_ALT_F5       }, ;
    {xbeK_ALT_F6        ,  K_ALT_F6       }, ;
    {xbeK_ALT_F7        ,  K_ALT_F7       }, ;
    {xbeK_ALT_F8        ,  K_ALT_F8       }, ;
    {xbeK_ALT_F9        ,  K_ALT_F9       }, ;
    {xbeK_ALT_F10       ,  K_ALT_F10      }, ;
    {xbeK_ALT_F11       ,  K_ALT_F11      }, ;
    {xbeK_ALT_F12       ,  K_ALT_F12      }, ;
    {xbeK_SH_F1         ,  K_SH_F1        }, ;
    {xbeK_SH_F2         ,  K_SH_F2        }, ;
    {xbeK_SH_F3         ,  K_SH_F3        }, ;
    {xbeK_SH_F4         ,  K_SH_F4        }, ;
    {xbeK_SH_F5         ,  K_SH_F5        }, ;
    {xbeK_SH_F6         ,  K_SH_F6        }, ;
    {xbeK_SH_F7         ,  K_SH_F7        }, ;
    {xbeK_SH_F8         ,  K_SH_F8        }, ;
    {xbeK_SH_F9         ,  K_SH_F9        }, ;
    {xbeK_SH_F10        ,  K_SH_F10       }, ;
    {xbeK_SH_F11        ,  K_SH_F11       }, ;
    {xbeK_SH_F12        ,  K_SH_F12       }  }

    nPtr :=  Ascan(KeyCodes, {|elem| nKey = elem[1]})
    if nPtr > 0
       nKey := KeyCodes[nPtr][2]
    endif

Return nKey
//////////////////////////////////////////////////////////////////////

